@@ -1,5 +1,5 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 2011-2024 Free Software Foundation, Inc.
+   Copyright (C) 2011-2025 Free Software Foundation, Inc.
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
@@ -126,6 +126,10 @@ mu_cidr_from_string (struct mu_cidr *pcidr, const char *str)
   else
     masklen_to_netmask (cidr.netmask, cidr.len, cidr.len * 8);
 
+  /* Mask out extra bits, to ensure correct first IP. */
+  for (len = cidr.len; len > 0 && cidr.netmask[len-1] != 0xff; len--)
+    cidr.address[len-1] &= cidr.netmask[len-1];
+
   memcpy (pcidr, &cidr, sizeof (*pcidr));
   return 0;
 }
