@@ -1,5 +1,5 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 2003-2022 Free Software Foundation, Inc.
+   Copyright (C) 2003-2025 Free Software Foundation, Inc.
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
@@ -199,7 +199,7 @@ _dot_encoder (void *xd,
 	    {
 	      if (i == 0)
 		{
-		  iobuf->osize = 2;
+		  iobuf->osize += 2;
 		  return mu_filter_moreoutput;
 		}
 	      break;
