@@ -63,7 +63,7 @@ percent_encoder (void *xd,
 	    {
 	      if (i == 0)
 		{
-		  iobuf->osize = 3;
+		  iobuf->osize += 3;
 		  return mu_filter_moreoutput;
 		}
 	      break;
