@@ -1,5 +1,5 @@
 /* wsp - test program for wordsplit
-   Copyright (C) 2014-2019 Sergey Poznyakoff
+   Copyright (C) 2014-2023 Sergey Poznyakoff
 
    Wordsplit is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
@@ -22,7 +22,11 @@
 #include <string.h>
 #include <assert.h>
 #include <errno.h>
-#include "wordsplit-version.h"
+#ifndef WORDSPLIT_VERSION
+# ifdef HAVE_WORDSPLIT_VERSION_H
+#  include "wordsplit-version.h"
+# endif
+#endif
 #include "wordsplit.h"
 
 extern char **environ;
@@ -33,7 +37,7 @@ char *progname;
 enum
   {
     TRIMNL_OPTION    = 0x01,   /* Remove trailing newline */
-    PLAINTEXT_OPTION = 0x02    /* Print input verbatim (no escapes) */
+    PLAINTEXT_OPTION = 0x02    /* Print intput verbatim (no escapes) */
   };
 
 /* Environment types */
@@ -96,6 +100,10 @@ struct wsopt
 /* Index of the next argument in the argv */
 static int wsoptind = -1;
 
+#ifndef WORDSPLIT_VERSION
+# error "WORDSPLIT_VERSION not defined"
+#endif
+
 void
 print_version (void)
 {
@@ -213,6 +221,8 @@ getwsopt (int argc, char **argv, struct wsopt *wso, struct wsclosure *wsc)
 		}
 	      arg = argv[wsoptind++];
 	    }
+          else
+            arg = NULL;
 	  wso->setfn (wso->tok, negate, arg, wsc);
 	}
       return 0;
