@@ -1,5 +1,5 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 1999-2024 Free Software Foundation, Inc.
+   Copyright (C) 1999-2025 Free Software Foundation, Inc.
 
    GNU Mailutils is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -315,6 +315,7 @@ static char *mail_capa[] = {
   "debug",
   "mailbox",
   "locking",
+  "mailer",
   NULL
 };
 
@@ -475,30 +476,14 @@ main (int argc, char **argv)
   setenv ("VISUAL", "vi", 0);
 
   /* set defaults for execution */
-  util_do_command ("setq PID=\"%lu\"", (unsigned long) getpid ());
   for (i = 0; i < sizeof (default_setup)/sizeof (default_setup[0]); i++)
     util_do_command ("%s", default_setup[i]);
 
-  p = getenv ("LINES");
-  if (p && p[strspn (p, "0123456789")] == 0)
-    util_do_command ("set screen=%s", p);
-  else
-    util_do_command ("set screen=%d", util_getlines ());
-
-  p = getenv ("COLUMNS");
-  if (p && p[strspn (p, "0123456789")] == 0)
-    util_do_command ("set columns=%s", p);
-  else
-    util_do_command ("set columns=%d", util_getcols ());
-
-  /* Set the default mailer to sendmail. FIXME: Minor memory leak. */
-  mailvar_set (mailvar_name_sendmail,
-	       mu_strdup ("sendmail:" PATH_SENDMAIL), mailvar_type_string,
-	       MOPTF_OVERWRITE);
-
-  /* argument parsing */
+  /* Parse command line arguments and mailutils-style configuration. */
   mu_cli (argc, argv, &cli, mail_capa, NULL, &argc, &argv);
 
+  mailvar_init_variables ();
+
   if (default_content_type || default_encoding)
     mime_option = 1;
   if (mime_option)
