@@ -24,19 +24,19 @@
 #include <pwd.h>
 #include <grp.h>
 #include <unistd.h>
+#include <fcntl.h>
 #include <mailutils/mailutils.h>
 #include <mailutils/tls.h>
 #include "mailutils/cli.h"
+#include "mailutils/md5.h"
 #include <muaux.h>
 #ifdef HAVE_TERMIOS_H
 # include <termios.h>
 #endif
 #include <sys/ioctl.h>
 
-static int reverse_order;
-static int preserve_mail; 
+static int preserve_mail;
 static int emacs_mode;
-static int uidl_option;
 static int verbose_option;
 static int ignore_errors;
 static char *program_id_option;
@@ -44,15 +44,33 @@ static size_t max_messages_option;
 static int notify;
 static int progress_meter_option;
 
+static char *source_name, *dest_name;
+static mu_mailbox_t source, dest;
+
+/* Synchronisation modes. */
+enum sync_mode
+  {
+    SYNC_ALL,     /* Download all messages. */
+    SYNC_UIDNEXT, /* Use uidvalidity/uidnext to select messages (IMAP) */
+    SYNC_UIDL     /* Use UIDL to select messages. */
+  };
+
+enum sync_mode sync_mode;
+
+#ifndef DEFAULT_MOVEMAIL_SYNC_DIR
+# define DEFAULT_MOVEMAIL_SYNC_DIR "~/.movemail.sync"
+#endif
+static char *sync_dir = DEFAULT_MOVEMAIL_SYNC_DIR;
+
   /* These bits tell what to do when an error occurs: */
 #define ONERROR_SKIP     0x01  /* Skip to the next message */
 #define ONERROR_DELETE   0x02  /* Delete the source message */
-#define ONERROR_COUNT    0x04  /* Count it as processed */ 
+#define ONERROR_COUNT    0x04  /* Count it as processed */
 
-static int onerror_flags;  
+static int onerror_flags;
 
 size_t msg_count = 0;         /* Number of processed messages */
-size_t get_err_count = 0;     /* Number of message retrieval errors */ 
+size_t get_err_count = 0;     /* Number of message retrieval errors */
 size_t app_err_count = 0;     /* Number of message appending errors */
 
 enum set_ownership_mode
@@ -95,14 +113,37 @@ get_next_so_method ()
 }
 
 mu_kwd_t method_kwd[] = {
-  { "copy-id", copy_owner_id }, 
+  { "copy-id", copy_owner_id },
   { "copy-name", copy_owner_name },
   { "set-name", set_owner_name },
   { "user", set_owner_name },
   { "set-id", set_owner_id },
   { NULL }
 };
+
+mu_kwd_t sync_kwd[] = {
+  { "all",      SYNC_ALL },
+  { "uidnext",  SYNC_UIDNEXT },
+  { "imap",     SYNC_UIDNEXT },
+  { "uidl",     SYNC_UIDL },
+  { "pop",      SYNC_UIDL },
+  { NULL }
+};
 
+static int
+set_sync_mode (const char *str)
+{
+  int n;
+
+  if (mu_kwd_xlat_name (sync_kwd, str, &n))
+    {
+      mu_error (_("invalid synchronization method: %s"), str);
+      return 1;
+    }
+  sync_mode = n;
+  return 0;
+}
+
 static int
 set_mailbox_ownership (const char *str)
 {
@@ -114,13 +155,13 @@ set_mailbox_ownership (const char *str)
       char *p;
       size_t len = strcspn (str, "=");
       struct set_ownership_method *meth;
-	  
+
       if (mu_kwd_xlat_name_len (method_kwd, str, len, &code))
 	{
 	  mu_error (_("invalid ownership method: %s"), str);
 	  return 1;
 	}
-      
+
       meth = get_next_so_method ();
       meth->mode = code;
       switch (meth->mode)
@@ -128,7 +169,7 @@ set_mailbox_ownership (const char *str)
 	case copy_owner_id:
 	case copy_owner_name:
 	  break;
-	  
+
 	case set_owner_id:
 	  if (!str[len])
 	    {
@@ -158,7 +199,7 @@ set_mailbox_ownership (const char *str)
 	  else
 	    meth->owner.id.gid = (gid_t) -1;
 	  break;
-	  
+
 	case set_owner_name:
 	  if (!str[len])
 	    {
@@ -180,7 +221,7 @@ set_mailbox_ownership_list (char const *str)
     {
       struct mu_wordsplit ws;
       size_t i;
-      
+
       ws.ws_delim = ",";
       if (mu_wordsplit (str, &ws, MU_WRDSF_DEFFLAGS|MU_WRDSF_DELIM))
 	{
@@ -237,7 +278,7 @@ set_onerror_actions (char const *str)
 {
   mu_list_t list;
   int rc;
-  
+
   mu_list_create (&list);
   mu_list_set_destroy_item (list, mu_list_free_item);
   mu_string_split (str, ",", list);
@@ -246,6 +287,19 @@ set_onerror_actions (char const *str)
   return rc;
 }
 
+static void
+cli_sync_mode (struct mu_parseopt *po, struct mu_option *opt, char const *arg)
+{
+  if (set_sync_mode (arg))
+    exit (po->po_exit_error);
+}
+
+static void
+cli_uidl (struct mu_parseopt *po, struct mu_option *opt, char const *arg)
+{
+  sync_mode = SYNC_UIDL;
+}
+
 static void
 cli_mailbox_ownership (struct mu_parseopt *po, struct mu_option *opt,
 		       char const *arg)
@@ -266,43 +320,47 @@ static struct mu_option movemail_options[] = {
     N_("preserve the source mailbox"),
     mu_c_bool, &preserve_mail },
   { "keep-messages", 0, NULL, MU_OPTION_ALIAS },
-  
-  { "reverse",  'r', NULL, MU_OPTION_DEFAULT,
-    N_("reverse the sorting order"),
-    mu_c_bool, &reverse_order },
-  
+
   { "emacs",      0, NULL, MU_OPTION_DEFAULT,
     N_("output information used by Emacs rmail interface"),
     mu_c_bool, &emacs_mode },
-  
+
+  { "sync",     's', "all|uidnext|uidl", MU_OPTION_DEFAULT,
+    N_("set mail synchronization mode"),
+    mu_c_string, NULL, cli_sync_mode },
+
   { "uidl",     'u', NULL, MU_OPTION_DEFAULT,
-    N_("use UIDLs to avoid downloading the same message twice"),
-    mu_c_bool, &uidl_option },
-  
+    N_("same as --sync=uidl"),
+    mu_c_string, NULL, cli_uidl },
+
+  { "sync-dir", 0, N_("DIR"), MU_OPTION_DEFAULT,
+    N_("Set synchronization directory"),
+    mu_c_string, &sync_dir },
+
   { "verbose",  'v', NULL, MU_OPTION_DEFAULT,
     N_("increase verbosity level"),
     mu_c_incr, &verbose_option },
-  
+
   { "owner",    'P', N_("MODELIST"), MU_OPTION_DEFAULT,
     N_("control mailbox ownership"),
     mu_c_string, NULL, cli_mailbox_ownership },
-  
+
   { "ignore-errors", 0, NULL, MU_OPTION_DEFAULT,
     N_("try to continue after errors"),
     mu_c_bool, &ignore_errors },
-  
+
   { "onerror",       0, N_("KW[,KW...]"), MU_OPTION_DEFAULT,
     N_("what to do on errors"),
     mu_c_string, NULL, cli_onerror },
-  
+
   { "program-id",    0, N_("FMT"), MU_OPTION_DEFAULT,
     N_("set program identifier for diagnostics (default: program name)"),
     mu_c_string, &program_id_option },
-  
+
   { "max-messages",  0, N_("NUMBER"), MU_OPTION_DEFAULT,
     N_("process at most NUMBER messages"),
     mu_c_size, &max_messages_option },
-  
+
   { "notify",        0, NULL,   MU_OPTION_DEFAULT,
     N_("enable biff notification"),
     mu_c_bool, &notify },
@@ -314,14 +372,39 @@ static struct mu_option movemail_options[] = {
   MU_OPTION_END
 }, *options[] = { movemail_options, NULL };
 
+static int
+cb_sync_mode (void *data, mu_config_value_t *val)
+{
+  if (mu_cfg_assert_value_type (val, MU_CFG_STRING))
+    return 1;
+  return set_sync_mode (val->v.string);
+}
+
+static int
+cb_uidl (void *data, mu_config_value_t *val)
+{
+  int b;
+
+  if (mu_cfg_assert_value_type (val, MU_CFG_STRING))
+    return 1;
+  if (mu_str_to_c (val->v.string, mu_c_bool, &b, NULL))
+    {
+      mu_error (_("not a boolean"));
+      return 1;
+    }
+  if (b)
+    sync_mode = SYNC_UIDL;
+  return 0;
+}
+
 static int
 cb_mailbox_ownership (void *data, mu_config_value_t *val)
 {
   int i;
-  
+
   if (val->type == MU_CFG_STRING)
     set_mailbox_ownership_list (val->v.string);
-		
+
   if (mu_cfg_assert_value_type (val, MU_CFG_LIST))
     return 1;
 
@@ -353,16 +436,26 @@ cb_onerror (void *data, mu_config_value_t *val)
     }
   return 0;
 }
-  
+
 struct mu_cfg_param movemail_cfg_param[] = {
   { "preserve", mu_c_bool, &preserve_mail, 0, NULL,
     N_("Do not remove messages from the source mailbox.") },
-  { "reverse",  mu_c_bool, &reverse_order, 0, NULL,
-    N_("Reverse message sorting order.") },
   { "emacs", mu_c_bool, &emacs_mode, 0, NULL,
     N_("Output information used by Emacs rmail interface.") },
-  { "uidl", mu_c_bool, &uidl_option, 0, NULL,
-    N_("Use UIDLs to avoid downloading the same message twice.") },
+  { "sync", mu_c_string, NULL, 0, cb_sync_mode,
+    N_("Set synchronization mode. Argument is one of:\n"
+       " all     - download all messages (default);\n"
+       " uidnext - use uidvalidity/uidnext values to select messages to download;\n"
+       " imap    - same as uid;\n"
+       " uidl    - use UIDL to avoid downloading the same message twice;\n"
+       " pop     - same as uidl;"),
+    N_("mode") },
+  { "uidl", mu_c_bool, NULL, 0, cb_uidl,
+    N_("Same as \"sync uidl\".") },
+
+  { "sync-dir", mu_c_string, &sync_dir, 0, NULL,
+    N_("Set synchronization directory.") },
+
   { "verbose", mu_c_int, &verbose_option, 0, NULL,
     N_("Set verbosity level.") },
   { "program-id", mu_c_string, &program_id_option, 0, NULL,
@@ -404,14 +497,14 @@ static char *movemail_capa[] = {
   "locking",
   "mailbox",
   "auth",
-  NULL 
+  NULL
 };
 
 void
 die (mu_mailbox_t mbox, const char *msg, int status)
 {
   mu_url_t url = NULL;
-  
+
   mu_mailbox_get_url (mbox, &url);
   if (emacs_mode)
     mu_error (_("%s:mailbox `%s': %s: %s"),
@@ -430,11 +523,11 @@ lock_mailbox (mu_mailbox_t mbox)
 {
   mu_locker_t lock;
   int status;
-  
+
   status = mu_mailbox_get_locker (mbox, &lock);
   if (status)
     die (mbox, _("cannot retrieve locker"), status);
-      
+
   if (!lock)
     /* Remote mailboxes have no lockers */
     return;
@@ -461,7 +554,7 @@ attach_passwd_ticket (mu_mailbox_t mbx, char *passwd)
       mu_error ("mu_secret_create: %s", mu_strerror (rc));
       exit (1);
     }
-  
+
   mu_ticket_create (&t, NULL);
   mu_ticket_set_secret (t, secret);
 
@@ -541,7 +634,7 @@ movemail (mu_mailbox_t dst, mu_message_t msg, size_t msgno)
       else
 	return 1;
     }
-  return max_messages_option && msg_count >= max_messages_option;
+  return 0;
 }
 
 /* Open source mailbox using compatibility syntax. Source_name is
@@ -570,14 +663,12 @@ compatibility_mode (mu_mailbox_t *mbx, char *source_name, char *password,
   free (tmp);
 }
 
-static mu_mailbox_t source, dest;
-
 static void
 close_mailboxes (void)
 {
   mu_mailbox_close (dest);
   mu_mailbox_close (source);
-}  
+}
 
 static int
 get_mbox_owner_id (mu_mailbox_t mbox, mu_url_t url, struct user_id *id)
@@ -592,7 +683,7 @@ get_mbox_owner_id (mu_mailbox_t mbox, mu_url_t url, struct user_id *id)
        || strcmp (s, "maildir") == 0))
     {
       struct stat st;
-      
+
       rc = mu_url_sget_path  (url, &s);
       if (rc)
 	die (mbox, _("cannot get path"), rc);
@@ -615,7 +706,7 @@ static int
 get_user_id (const char *name, struct user_id *id)
 {
   struct mu_auth_data *auth = mu_get_auth_by_name (name);
-  
+
   if (!auth)
     {
       if (verbose_option)
@@ -627,7 +718,7 @@ get_user_id (const char *name, struct user_id *id)
   id->gid = auth->gid;
   mu_auth_data_free (auth);
   return 0;
-}  
+}
 
 static int
 get_mbox_owner_name (mu_mailbox_t mbox, mu_url_t url, struct user_id *id)
@@ -647,7 +738,7 @@ guess_mbox_owner (mu_mailbox_t mbox, struct user_id *id)
   mu_url_t url = NULL;
   int rc;
   struct set_ownership_method *meth;
-  
+
   rc = mu_mailbox_get_url (mbox, &url);
   if (rc)
     {
@@ -663,11 +754,11 @@ guess_mbox_owner (mu_mailbox_t mbox, struct user_id *id)
 	case copy_owner_id:
 	  rc = get_mbox_owner_id (mbox, url, id);
 	  break;
-	  
+
 	case copy_owner_name:
 	  rc = get_mbox_owner_name (mbox, url, id);
 	  break;
-	  
+
 	case set_owner_id:
 	  id->uid = meth->owner.id.uid;
 	  rc = 0;
@@ -688,13 +779,13 @@ guess_mbox_owner (mu_mailbox_t mbox, struct user_id *id)
 	  else
 	    id->gid = meth->owner.id.gid;
 	  break;
-	  
+
 	case set_owner_name:
 	  rc = get_user_id (meth->owner.name, id);
 	  break;
 	}
     }
-  
+
   return rc;
 }
 
@@ -713,7 +804,7 @@ switch_owner (mu_mailbox_t mbox)
 			_("ignoring mailbox-ownership statement"));
       return;
     }
-  
+
   if (guess_mbox_owner (mbox, &user_id) == 0)
     {
       if (mu_switch_to_privs (user_id.uid, user_id.gid, NULL))
@@ -750,7 +841,7 @@ static int
 get_url_part (mu_url_t url, const char *name, size_t nlen, char **ret)
 {
   int rc;
-  
+
   if (!url)
     return MU_WRDSE_UNDEF;
   if (SEQ ("user", name, nlen))
@@ -768,7 +859,7 @@ get_url_part (mu_url_t url, const char *name, size_t nlen, char **ret)
     {
     case 0:
       break;
-      
+
     case MU_ERR_NOENT:
       return MU_WRDSE_UNDEF;
 
@@ -777,10 +868,10 @@ get_url_part (mu_url_t url, const char *name, size_t nlen, char **ret)
 	return MU_WRDSE_NOSPACE;
       return MU_WRDSE_USERERR;
     }
-  
+
   return MU_WRDSE_OK;
 }
-    
+
 static int
 movemail_getvar (char **ret, const char *name, size_t nlen, void *data)
 {
@@ -792,7 +883,7 @@ movemail_getvar (char **ret, const char *name, size_t nlen, void *data)
 
   if (nlen > 5 && memcmp ("dest_", name, 5) == 0)
     return get_url_part (pc->dest_url, name + 5, nlen - 5, ret);
-  
+
   if (SEQ ("progname", name, nlen))
     s = mu_program_name;
   else if (SEQ ("source", name, nlen))
@@ -801,7 +892,7 @@ movemail_getvar (char **ret, const char *name, size_t nlen, void *data)
     s = pc->dest_name;
   else
     return MU_WRDSE_UNDEF;
-    
+
   *ret = strdup (s);
   if (!*ret)
     return MU_WRDSE_NOSPACE;
@@ -827,7 +918,7 @@ set_program_id (const char *source_name, const char *dest_name)
     mu_diag_output (MU_DIAG_INFO,
 		    _("cannot obtain destination mailbox URL: %s"),
 		    mu_strerror (rc));
-  
+
   ws.ws_getvar = movemail_getvar;
   ws.ws_closure = &clos;
   if (mu_wordsplit (program_id_option, &ws,
@@ -838,7 +929,7 @@ set_program_id (const char *source_name, const char *dest_name)
 		mu_wordsplit_strerror (&ws));
       return;
     }
-  
+
   /* FIXME: Don't use mu_set_program_name here, because it
      plays wise with its argument. We need a mu_set_diag_prefix
      function. */
@@ -868,7 +959,7 @@ static void
 progress_format (size_t pos, size_t count)
 {
   int n;
-  
+
   fputc ('\r', stdout);
   n = printf ("message %zu/%zu", pos, count);
   n = screen_width () - n;
@@ -881,7 +972,7 @@ void
 progress_start (mu_iterator_t itr)
 {
   size_t count;
-  
+
   if (!progress_meter_option)
     return;
 
@@ -897,7 +988,7 @@ void
 progress_mark (mu_iterator_t itr)
 {
   size_t count, pos;
-  
+
   if (!progress_meter_option)
     return;
 
@@ -907,8 +998,6 @@ progress_mark (mu_iterator_t itr)
       progress_meter_option = 0;
       return;
     }
-  if (reverse_order)
-    pos = count - pos + 1;
   progress_format (pos, count);
 }
 
@@ -919,25 +1008,459 @@ progress_stop (void)
     fputc ('\n', stdout);
 }
 
+static void
+sync_loop (mu_iterator_t itr)
+{
+  progress_start (itr);
+  for (mu_iterator_first (itr); !mu_iterator_is_done (itr);
+       mu_iterator_next (itr))
+    {
+      mu_message_t msg;
+      size_t msgno;
+      int rc;
+
+      rc = mu_iterator_ctl (itr, mu_itrctl_tell, &msgno);
+      if (rc)
+	{
+	  mu_error (_("cannot get iterator position: %s"),
+		    mu_strerror (rc));
+	  exit (1);
+	}
+
+      rc = mu_iterator_current (itr, (void **)&msg);
+      if (rc)
+	{
+	  mu_error (_("cannot read message %lu: %s"),
+		    (unsigned long) msgno, mu_strerror (rc));
+	  get_err_count++;
+	  continue;
+	}
+
+      if (movemail (dest, msg, msgno))
+	break;
+      progress_mark (itr);
+    }
+  progress_stop ();
+}
+
+static void
+movemail_sync_all (void)
+{
+  int rc;
+  mu_iterator_t itr;
+  size_t start = 1, end;
+
+  rc = mu_mailbox_messages_count (source, &end);
+  if (rc)
+    {
+      mu_error (_("%s: can't get number of messages: %s"),
+		source_name, mu_strerror (rc));
+      exit (1);
+    }
+
+  if (max_messages_option)
+    {
+      if (end > max_messages_option)
+	start = end - max_messages_option + 1;
+    }
+
+  rc = mu_mailbox_get_iterator_at (source, &itr, start, end, 0);
+  if (rc)
+    {
+      mu_error (_("cannot obtain mailbox iterator: %s"), mu_strerror (rc));
+      exit (1);
+    }
+
+  sync_loop (itr);
+  mu_iterator_destroy (&itr);
+}
+
+static void
+movemail_sync_uidl (void)
+{
+  int rc;
+  mu_list_t orig_uidl_list, src_uidl_list, dst_uidl_list;
+  mu_iterator_t itr;
+  size_t count;
+
+  rc = mu_mailbox_get_uidls (source, &orig_uidl_list);
+  if (rc)
+    die (source, _("cannot get UIDLs"), rc);
+
+  src_uidl_list = orig_uidl_list;
+  if (max_messages_option > 0)
+    {
+      rc = mu_list_count (orig_uidl_list, &count);
+      if (rc)
+	{
+	  mu_diag_funcall (MU_DIAG_ERROR, "mu_list_count", NULL, errno);
+	  exit (1);
+	}
+      if (count > max_messages_option)
+	{
+	  rc = mu_list_slice2 (&src_uidl_list, orig_uidl_list,
+			       count - max_messages_option,
+			       count);
+	  if (rc)
+	    {
+	      mu_diag_funcall (MU_DIAG_ERROR, "mu_list_slice2", NULL, errno);
+	      exit (1);
+	    }
+	}
+    }
+
+  rc = mu_mailbox_get_uidls (dest, &dst_uidl_list);
+  if (rc)
+    die (dest, _("cannot get UIDLs"), rc);
+
+  mu_list_set_comparator (dst_uidl_list, _compare_uidls);
+
+  mu_list_get_iterator (src_uidl_list, &itr);
+  for (mu_iterator_first (itr); !mu_iterator_is_done (itr);
+       mu_iterator_next (itr))
+    {
+      struct mu_uidl *uidl;
+
+      mu_iterator_current (itr, (void **)&uidl);
+      if (mu_list_locate (dst_uidl_list, uidl, NULL) == 0)
+	mu_iterator_ctl (itr, mu_itrctl_delete, NULL);
+    }
+  mu_list_destroy (&dst_uidl_list);
+  mu_list_set_comparator (src_uidl_list, NULL);
+
+  progress_start (itr);
+  for (mu_iterator_first (itr); !mu_iterator_is_done (itr);
+       mu_iterator_next (itr))
+    {
+      struct mu_uidl *uidl;
+      mu_message_t msg;
+      mu_header_t hdr;
+
+      mu_iterator_current (itr, (void **)&uidl);
+
+      if ((rc = mu_mailbox_get_message (source, uidl->msgno, &msg)) != 0)
+	{
+	  mu_error (_("cannot read message %lu: %s"),
+		    (unsigned long) uidl->msgno, mu_strerror (rc));
+	  get_err_count++;
+	  continue;
+	}
+
+      /* Check if the downloaded message has X-UIDL header. If not,
+	 add one. This check should disappear once mailutils implements
+	 alternative storage for mailbox meta-data. */
+      if ((rc = mu_message_get_header (msg, &hdr)))
+	{
+	  mu_error (_("%lu: cannot get header: %s"),
+		    (unsigned long) uidl->msgno, mu_strerror (rc));
+	}
+      else
+	{
+	  char const *suidl = NULL;
+
+	  if ((rc = mu_header_sget_value (hdr, MU_HEADER_X_UIDL, &suidl)))
+	    {
+	      if (rc != MU_ERR_NOENT)
+		mu_error (_("%lu: cannot get %s: %s"),
+			  (unsigned long) uidl->msgno, MU_HEADER_X_UIDL,
+			  mu_strerror (rc));
+	    }
+	  else if (strcmp (suidl, uidl->uidl))
+	    {
+	      mu_error (_("%lu: stored and reported UIDL differ; fixing"),
+			(unsigned long) uidl->msgno);
+	      suidl = NULL;
+	    }
+
+	  if ((rc = mu_header_set_value (hdr, MU_HEADER_X_UIDL,
+					 uidl->uidl, 1)))
+	    {
+	      mu_error (_("%lu: cannot set header: %s"),
+			(unsigned long) uidl->msgno, mu_strerror (rc));
+	    }
+	}
+      progress_mark (itr);
+      if (movemail (dest, msg, uidl->msgno))
+	break;
+    }
+  progress_stop ();
+  mu_iterator_destroy (&itr);
+  mu_list_destroy (&orig_uidl_list);
+}
+
+struct sync_prop
+{
+  char *filename;
+  FILE *fp;
+  mu_locker_t lck;
+  unsigned long uidvalidity;
+  unsigned long uidnext;
+  int modified;
+};
+
+static inline unsigned long
+sync_prop_get_uidvalidity (struct sync_prop *prop)
+{
+  return prop->uidvalidity;
+}
+
+static inline unsigned long
+sync_prop_get_uidnext (struct sync_prop *prop)
+{
+  return prop->uidnext;
+}
+
+static inline void
+sync_prop_update (struct sync_prop *prop, unsigned long uidvalidity,
+		  size_t uidnext)
+{
+  prop->uidvalidity = uidvalidity;
+  prop->uidnext = uidnext;
+  prop->modified++;
+}
+
+int
+sync_prop_save (struct sync_prop *prop)
+{
+  if (prop->modified)
+    {
+      rewind (prop->fp);
+      fflush (prop->fp);
+      ftruncate (fileno (prop->fp), 0);
+      fprintf (prop->fp, "%lu\n%lu\n", prop->uidvalidity, prop->uidnext);
+      if (ferror (prop->fp))
+	{
+	  mu_error (_("error writing to %s: %s"), prop->filename,
+		    mu_strerror (errno));
+	  return -1;
+	}
+      prop->modified = 0;
+    }
+  return 0;
+}
+
+void
+sync_prop_close (struct sync_prop *prop)
+{
+  if (prop->fp != NULL)
+    {
+      if (prop->modified)
+	sync_prop_save (prop);
+
+      fclose (prop->fp);
+      prop->fp = NULL;
+      mu_locker_unlock (prop->lck);
+      mu_locker_destroy (&prop->lck);
+      free (prop->filename);
+    }
+}
+
+void
+sync_prop_open (struct sync_prop *prop)
+{
+  int rc;
+  char *dir = mu_tilde_expansion (sync_dir, MU_HIERARCHY_DELIMITER, NULL);
+  struct stat st;
+  struct mu_md5_ctx md5context;
+  unsigned char md5digest[16];
+  char buf[2 * 16 + 1], *p;
+  size_t i;
+  char *statefile;
+  int fd;
+
+  if (stat (dir, &st))
+    {
+      if (errno == ENOENT)
+	{
+	  if (mkdir (dir, 0700))
+	    {
+	      mu_error (_("can't stat %s: %s"), dir, mu_strerror (errno));
+	      exit (1);
+	    }
+	}
+      else
+	{
+	  mu_error (_("can't stat %s: %s"), dir, mu_strerror (errno));
+	  exit (1);
+	}
+    }
+  else if (!S_ISDIR (st.st_mode))
+    {
+      mu_error (_("%s exists, but is not a directory"), dir);
+      exit (1);
+    }
+
+  mu_md5_init_ctx (&md5context);
+  mu_md5_process_bytes (source_name, strlen (source_name), &md5context);
+  mu_md5_process_bytes (dest_name, strlen (dest_name), &md5context);
+  mu_md5_finish_ctx (&md5context, md5digest);
+
+  for (i = 0, p = buf; i < sizeof (md5digest); i++, p += 2)
+    sprintf (p, "%02x", md5digest[i]);
+  *p = 0;
+
+  if ((statefile = mu_make_file_name (dir, buf)) == NULL)
+    {
+      mu_diag_funcall (MU_DIAG_ERROR, "mu_make_file_name", NULL, errno);
+      exit (1);
+    }
+
+  memset (prop, 0, sizeof (*prop));
+  prop->filename = statefile;
+
+  if ((rc = mu_locker_create_ext (&prop->lck, statefile, NULL)) != 0)
+    {
+      mu_diag_funcall (MU_DIAG_ERROR, "mu_locker_create_ext", NULL, errno);
+      exit (1);
+    }
+
+  if ((rc = mu_locker_lock (prop->lck)) != 0)
+    {
+      mu_error (_("can't lock %s: %s"), statefile, mu_strerror (rc));
+      exit (1);
+    }
+
+  fd = open (prop->filename, O_CREAT | O_RDWR, 0600);
+  if (fd == -1)
+    {
+      mu_locker_unlock (prop->lck);
+      mu_error (_("can't open %s: %s"), prop->filename, mu_strerror (errno));
+      exit (1);
+    }
+
+  prop->fp = fdopen (fd, "r+");
+  if (fscanf (prop->fp, "%lu\n%lu", &prop->uidvalidity, &prop->uidnext) != 2)
+    {
+      prop->uidvalidity = 0;
+      prop->uidnext = 0;
+    }
+}
+
+static void
+movemail_sync_uidnext (void)
+{
+  struct sync_prop prop;
+  unsigned long uidvalidity = 0;
+  size_t uidstart = 1;
+  size_t uidstop = 0;
+  size_t uidnext;
+  int rc;
+  mu_iterator_t itr;
+
+  if ((rc = mu_mailbox_uidvalidity (source, &uidvalidity)) != 0)
+    {
+      mu_error (_("%s: can't get uidvalidity: %s"), source_name,
+		mu_strerror (rc));
+      exit (1);
+    }
+  if ((rc = mu_mailbox_uidnext (source, &uidnext)) != 0)
+    {
+      mu_error (_("%s: can't get uidnext: %s"), source_name,
+		mu_strerror (rc));
+      exit (1);
+    }
+  uidstop = uidnext;
+
+  sync_prop_open (&prop);
+
+  if (uidvalidity == sync_prop_get_uidvalidity (&prop))
+    {
+      uidstart = sync_prop_get_uidnext (&prop);
+    }
+  else
+    {
+      uidstop = 0;
+      if (verbose_option)
+	mu_diag_output (MU_DIAG_WARNING,
+			_("%s: uidvalidity changed (%lu > %lu)"),
+			source_name, sync_prop_get_uidvalidity (&prop),
+			uidvalidity);
+    }
+
+  if (verbose_option)
+    {
+      if (uidstart == uidstop)
+	{
+	  mu_diag_output (MU_DIAG_WARNING, _("%s: no new messages"),
+			  source_name);
+	  sync_prop_close (&prop);
+	  return;
+	}
+      else
+	mu_diag_output (MU_DIAG_WARNING,
+			_("%s: downloading messages from UID %zu"),
+			source_name, uidstart);
+    }
+
+  rc = mu_mailbox_translate (source, MU_MAILBOX_UID_TO_MSGNO, uidstart,
+			     &uidstart);
+  if (rc)
+    {
+      if (verbose_option)
+	mu_diag_output (MU_DIAG_WARNING,
+			_("%s: start UID %zu not found: %s"),
+			source_name, uidstart, mu_strerror (rc));
+      uidstart = 1;
+    }
+
+  rc = mu_mailbox_messages_count (source, &uidstop);
+  if (rc)
+    {
+      mu_error (_("%s: can't get number of messages: %s"),
+		source_name, mu_strerror (rc));
+      sync_prop_close (&prop);
+      exit (1);
+    }
+
+  if (max_messages_option > 0)
+    {
+      if (uidstop - uidstart + 1 > max_messages_option)
+	uidstart = uidstop - max_messages_option + 1;
+    }
+
+  if (verbose_option)
+    mu_diag_output (MU_DIAG_WARNING,
+		    _("%s: downloading messages %zu to %zu"),
+		    source_name, uidstart, uidstop);
+
+  rc = mu_mailbox_get_iterator_at (source, &itr, uidstart, uidstop, 0);
+  if (rc)
+    {
+      mu_diag_funcall (MU_DIAG_ERROR, "mu_mailbox_get_iterator_at", NULL, rc);
+    }
+  else
+    {
+      sync_loop (itr);
+      sync_prop_update (&prop, uidvalidity, uidnext);
+    }
+  
+  sync_prop_close (&prop);
+
+  mu_iterator_destroy (&itr);
+}
+
+static void (*movemail_sync[])(void) = {
+  movemail_sync_all,
+  movemail_sync_uidnext,
+  movemail_sync_uidl,
+};
+
 int
 main (int argc, char **argv)
 {
   size_t total;
   int rc = 0;
-  char *source_name, *dest_name;
   int flags;
-  mu_list_t src_uidl_list = NULL;
-  mu_iterator_t itr;
-  
+
   /* Native Language Support */
   MU_APP_INIT_NLS ();
   MU_AUTH_REGISTER_ALL_MODULES ();
-  
+
   /* Register the desired "mailbox" formats.  */
   mu_register_all_formats ();
   /* Register authentication modules */
   mu_auth_register_module (&mu_auth_tls_module);
-  
+
   /* argument parsing */
   mu_cli (argc, argv, &cli, movemail_capa, NULL, &argc, &argv);
 
@@ -952,29 +1475,29 @@ main (int argc, char **argv)
 
   if (!isatty (1))
     progress_meter_option = 0;
-  
+
   if (emacs_mode)
-    {      
+    {
       /* Undo the effect of configuration options that may affect
 	 interaction with Emacs. */
       mu_registrar_set_default_record (mu_mbox_record);
       mu_stdstream_strerr_setup (MU_STRERR_STDERR);
     }
-  
+
   atexit (close_mailboxes);
-  
+
   source_name = argv[0];
   dest_name = argv[1];
 
   flags = preserve_mail ? MU_STREAM_READ : MU_STREAM_RDWR;
-  
+
   if (strncmp (source_name, "po:", 3) == 0)
     compatibility_mode (&source, source_name, argv[2], flags);
   else
     open_mailbox (&source, source_name, flags, argv[2]);
 
   switch_owner (source);
-  
+
   open_mailbox (&dest, dest_name,
 		MU_STREAM_APPEND | MU_STREAM_READ | MU_STREAM_CREAT, NULL);
 
@@ -988,14 +1511,14 @@ main (int argc, char **argv)
 	mu_error (_("failed to set up notification: %s"),
 		  mu_strerror (rc));
     }
-  
+
   rc = mu_mailbox_messages_count (source, &total);
   if (rc)
     {
       mu_error(_("cannot count messages: %s"), mu_strerror (rc));
       exit (1);
     }
-  
+
   if (verbose_option)
     {
       mu_diag_output (MU_DIAG_INFO,
@@ -1003,154 +1526,14 @@ main (int argc, char **argv)
 		      (unsigned long) total);
       if (max_messages_option)
 	mu_diag_output (MU_DIAG_INFO,
-			reverse_order ?
 			ngettext ("will process last %lu message",
 				  "will process last %lu messages",
-				  max_messages_option) :
-			ngettext ("will process first %lu message",
-				  "will process first %lu messages",
 				  max_messages_option),
 			(unsigned long) max_messages_option);
     }
-  
-  if (uidl_option)
-    {
-      mu_list_t dst_uidl_list = NULL;
 
-      rc = mu_mailbox_get_uidls (source, &src_uidl_list);
-      if (rc)
-	die (source, _("cannot get UIDLs"), rc);
-      rc = mu_mailbox_get_uidls (dest, &dst_uidl_list);
-      if (rc)
-	die (dest, _("cannot get UIDLs"), rc);
+  movemail_sync[sync_mode] ();
 
-      mu_list_set_comparator (dst_uidl_list, _compare_uidls);
-      
-      mu_list_get_iterator (src_uidl_list, &itr);
-      for (mu_iterator_first (itr); !mu_iterator_is_done (itr);
-	   mu_iterator_next (itr))
-	{
-	  struct mu_uidl *uidl;
-	      
-	  mu_iterator_current (itr, (void **)&uidl);
-	  if (mu_list_locate (dst_uidl_list, uidl, NULL) == 0)
-	    mu_iterator_ctl (itr, mu_itrctl_delete, NULL);
-	}
-      mu_list_destroy (&dst_uidl_list);
-      mu_list_set_comparator (src_uidl_list, NULL);
-
-      rc = mu_iterator_ctl (itr, mu_itrctl_set_direction, &reverse_order);
-      if (rc)
-	{
-	  mu_error (_("cannot set iteration direction: %s"), mu_strerror (rc));
-	  exit (1);
-	}
-      
-      progress_start (itr);
-      for (mu_iterator_first (itr); !mu_iterator_is_done (itr);
-	   mu_iterator_next (itr))
-	{
-	  struct mu_uidl *uidl;
-	  mu_message_t msg;
-	  mu_header_t hdr;
-
-	  mu_iterator_current (itr, (void **)&uidl);
-
-	  if ((rc = mu_mailbox_get_message (source, uidl->msgno, &msg)) != 0)
-	    {
-	      mu_error (_("cannot read message %lu: %s"),
-			(unsigned long) uidl->msgno, mu_strerror (rc));
-	      get_err_count++;
-	      continue;
-	    }
-
-	  /* Check if the downloaded message has X-UIDL header. If not,
-	     add one. This check should disappear one mailutils implements
-	     alternative storage for mailbox meta-data. */
-	  if ((rc = mu_message_get_header (msg, &hdr)))
-	    {
-	      mu_error (_("%lu: cannot get header: %s"),
-			(unsigned long) uidl->msgno, mu_strerror (rc));
-	    }
-	  else
-	    {
-	      char const *suidl = NULL;
-
-	      if ((rc = mu_header_sget_value (hdr, MU_HEADER_X_UIDL, &suidl)))
-		{
-		  if (rc != MU_ERR_NOENT)
-		    mu_error (_("%lu: cannot get %s: %s"),
-			      (unsigned long) uidl->msgno, MU_HEADER_X_UIDL,
-			      mu_strerror (rc));
-		}
-	      else if (strcmp (suidl, uidl->uidl))
-		{
-		  mu_error (_("%lu: stored and reported UIDL differ; fixing"),
-			    (unsigned long) uidl->msgno);
-		  suidl = NULL;
-		}
-
-	      if ((rc = mu_header_set_value (hdr, MU_HEADER_X_UIDL,
-					     uidl->uidl, 1)))
-		{
-		  mu_error (_("%lu: cannot set header: %s"),
-			    (unsigned long) uidl->msgno, mu_strerror (rc));
-		}
-	    }
-	  progress_mark (itr);
-	  if (movemail (dest, msg, uidl->msgno))
-	    break;
-	}
-    }
-  else
-    {
-      rc = mu_mailbox_get_iterator (source, &itr);
-      if (rc)
-	{
-	  mu_error (_("cannot obtain mailbox iterator: %s"), mu_strerror (rc));
-	  return 1;
-	}
-  
-      rc = mu_iterator_ctl (itr, mu_itrctl_set_direction, &reverse_order);
-      if (rc)
-	{
-	  mu_error (_("cannot set iteration direction: %s"),
-		    mu_strerror (rc));
-	  return 1;
-	}
-
-      progress_start (itr);
-      for (mu_iterator_first (itr); !mu_iterator_is_done (itr);
-	   mu_iterator_next (itr))
-	{
-	  mu_message_t msg;
-	  size_t msgno;
-	  
-	  rc = mu_iterator_ctl (itr, mu_itrctl_tell, &msgno);
-	  if (rc)
-	    {
-	      mu_error (_("cannot get iterator position: %s"),
-			mu_strerror (rc));
-	      return 1;
-	    }
-
-	  rc = mu_iterator_current (itr, (void **)&msg);
-	  if (rc)
-	    {
-	      mu_error (_("cannot read message %lu: %s"),
-			(unsigned long) msgno, mu_strerror (rc));
-	      get_err_count++;
-	      continue;
-	    }
-	  
-	  if (movemail (dest, msg, msgno))
-	    break;
-	  progress_mark (itr);
-	}
-    }
-  progress_stop ();
-  mu_iterator_destroy (&itr);
-  
   if (verbose_option)
     {
       mu_diag_output (MU_DIAG_INFO,
@@ -1161,12 +1544,12 @@ main (int argc, char **argv)
 		      (unsigned long) get_err_count,
 		      (unsigned long) app_err_count);
     }
-  
+
   if (app_err_count && !(onerror_flags & (ONERROR_DELETE|ONERROR_COUNT)))
     preserve_mail = 1;
   if (onerror_flags & ONERROR_COUNT)
     app_err_count = 0;
-      
+
   mu_mailbox_sync (dest);
   rc = mu_mailbox_close (dest);
   mu_mailbox_destroy (&dest);
@@ -1177,6 +1560,6 @@ main (int argc, char **argv)
 
   mu_mailbox_close (source);
   mu_mailbox_destroy (&source);
-  
+
   return !(rc == 0 && (app_err_count + get_err_count) == 0);
 }
