@@ -1,5 +1,5 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 2010-2020 Free Software Foundation, Inc.
+   Copyright (C) 2010-2025 Free Software Foundation, Inc.
 
    GNU Mailutils is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -456,8 +456,8 @@ shell_history (int argc, char **argv)
 char *
 readline (char *prompt)
 {
-  static size_t size = 0;
-  static char *buf = NULL;
+  size_t size = 0;
+  char *buf = NULL;
   size_t n;
 
   if (prompt)
@@ -469,8 +469,6 @@ readline (char *prompt)
     {
       free (buf);
       buf = NULL;
-      size = 0;
-      return NULL;
     }
   return buf;
 }
